;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;
;;; elephant.lisp -- package definition
;;; 
;;; Initial version 8/26/2004 by Ben Lee
;;; <blee@common-lisp.net>
;;; 
;;; part of
;;;
;;; Elephant: an object-oriented database for Common Lisp
;;;
;;; Copyright (c) 2004 by Andrew Blumberg and Ben Lee
;;; <ablumberg@common-lisp.net> <blee@common-lisp.net>
;;;
;;; This program is released under the following license
;;; ("GPL").  For differenct licensing terms, contact the
;;; copyright holders.
;;;
;;; This program is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU General
;;; Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at
;;; your option) any later version.
;;;
;;; This program is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE. See the GNU General Public License
;;; for more details.
;;;
;;; The GNU General Public License can be found in the file
;;; LICENSE which should have been distributed with this
;;; code.  It can also be found at
;;;
;;; http://www.opensource.org/licenses/gpl-license.php
;;;
;;; You should have received a copy of the GNU General
;;; Public License along with this program; if not, write
;;; to the Free Software Foundation, Inc., 59 Temple Place,
;;; Suite 330, Boston, MA 02111-1307 USA
;;;

(defpackage elephant
  (:documentation 
   "Elephant: an object-oriented database for Common Lisp.
Uses the SLEEPYCAT package to talk to Berkeley DB /
Sleepycat.")
  (:nicknames ele :ele)
  (:use common-lisp sleepycat uffi)
  (:shadow #:with-transaction)
  (:export #:*store-controller* #:*current-transaction* #:*auto-commit*
	   #:open-store #:close-store #:with-open-store
	   #:store-controller #:open-controller #:close-controller 
	   #:with-open-controller #:controller-path #:controller-environment
	   #:controller-db #:controller-root 
	   #:add-to-root #:get-from-root #:remove-from-root
	   #:start-transaction #:commit-transaction #:abort-transaction

	   #:persistent #:persistent-object #:persistent-metaclass

	   #:persistent-collection #:btree #:get-value #:remove-kv
	   #:indexed-btree #:add-index #:get-index #:remove-index
	   #:btree-index #:get-primary-key
	   #:indices #:primary #:key-form #:key-fn

	   #:cursor #:secondary-cursor #:make-cursor
	   #:with-btree-cursor #:map-btree #:cursor-close
	   #:cursor-duplicate #:cursor-current #:cursor-first
	   #:cursor-last #:cursor-next #:cursor-next-dup
	   #:cursor-next-nodup #:cursor-prev #:cursor-prev-nodup
	   #:cursor-set #:cursor-set-range #:cursor-get-both
	   #:cursor-get-both-range #:cursor-delete #:cursor-put
	   #:cursor-pcurrent #:cursor-pfirst #:cursor-plast
	   #:cursor-pnext #:cursor-pnext-dup #:cursor-pnext-nodup
	   #:cursor-pprev #:cursor-pprev-nodup #:cursor-pset
	   #:cursor-pset-range #:cursor-pget-both
	   #:cursor-pget-both-range

	   #:db-transaction-begin #:db-transaction-abort 
	   #:db-transaction-commit #:with-transaction
	   #:db-env-set-lock-detect #:db-env-get-lock-detect
	   #:db-transaction-id #:db-env-lock-id #:db-env-lock-id-free
	   #:db-env-lock-get #:db-env-lock-put #:with-lock
	   #:db-env-set-timeout #:db-env-get-timeout
	   #:db-env-set-flags #:db-env-get-flags
	   #:run-elephant-thread
	   )
  #+cmu  
  (:import-from :pcl
		compute-class-precedence-list
		validate-superclass
		standard-slot-definition
		standard-direct-slot-definition
		standard-effective-slot-definition
		direct-slot-definition-class
		effective-slot-definition-class
		slot-definition-name
		slot-definition-initform
		slot-definition-initfunction
		compute-effective-slot-definition
		class-slots
		slot-value-using-class
		slot-boundp-using-class
		slot-makunbound-using-class
		slot-definition-allocation
		slot-definition-initargs
		compute-slots

		initialize-internal-slot-functions
		compute-effective-slot-definition-initargs
		slot-definition-reader-function
		slot-definition-writer-function
		slot-definition-boundp-function
		slot-definition-allocation-class
		class-slot-cells
		plist-value
		+slot-unbound+) 
  #+cmu  
  (:import-from :ext
		make-weak-pointer weak-pointer-value finalize)

  #+cmu  
  (:import-from :bignum
		%bignum-ref)

  #+sbcl 
  (:import-from :sb-mop 
		compute-class-precedence-list
		validate-superclass
		standard-slot-definition
		standard-direct-slot-definition
		standard-effective-slot-definition
		direct-slot-definition-class
		effective-slot-definition-class
		slot-definition-name
		slot-definition-initform
		slot-definition-initfunction
		compute-effective-slot-definition
		class-slots
		slot-value-using-class
		slot-boundp-using-class
		slot-makunbound-using-class
		slot-definition-allocation
		slot-definition-initargs
		compute-slots)                                
  #+sbcl
  (:import-from :sb-pcl
		initialize-internal-slot-functions
		compute-effective-slot-definition-initargs
		slot-definition-reader-function
		slot-definition-writer-function
		slot-definition-boundp-function
		slot-definition-allocation-class
		class-slot-cells
		plist-value
		+slot-unbound+)
  #+sbcl
  (:import-from :sb-ext
		make-weak-pointer weak-pointer-value finalize)

  #+sbcl
  (:import-from :sb-bignum
		%bignum-ref)

  #+allegro
  (:import-from :clos
		compute-class-precedence-list
		validate-superclass
		standard-slot-definition
		standard-direct-slot-definition
		standard-effective-slot-definition
		direct-slot-definition-class
		effective-slot-definition-class
		slot-definition-name
		slot-definition-initform
		slot-definition-initfunction
		compute-effective-slot-definition
		class-slots
		slot-value-using-class
		slot-boundp-using-class
		slot-makunbound-using-class
		slot-definition-allocation
		slot-definition-initargs
		compute-slots)
  #+allegro
  (:import-from :excl
		compute-effective-slot-definition-initargs)
  #+openmcl
  (:import-from :ccl
		compute-class-precedence-list
		validate-superclass
		standard-slot-definition
		standard-direct-slot-definition
		standard-effective-slot-definition
		direct-slot-definition-class
		effective-slot-definition-class
		slot-definition-name
		slot-definition-initform
		slot-definition-initfunction
		compute-effective-slot-definition
		class-slots
		slot-value-using-class
		slot-boundp-using-class
		slot-makunbound-using-class
		slot-definition-allocation
		slot-definition-initargs
		compute-slots
		;; This stuff we need until we resolve the :transient
		;; slot specifier stuff
		make-effective-slot-definition
		slots-class
		%slot-definition-initfunction
		%slot-definition-documentation
		%slot-definition-initargs
		%slot-definition-initform
		%slot-definition-allocation
		%slot-definition-class
		%slot-definition-type)
  #+lispworks  
  (:import-from :clos
		compute-class-precedence-list
		validate-superclass
		standard-slot-definition
		standard-direct-slot-definition
		standard-effective-slot-definition
		direct-slot-definition-class
		effective-slot-definition-class
		slot-definition-name
		slot-definition-initform
		slot-definition-initfunction
		compute-effective-slot-definition
		class-slots
		slot-value-using-class
		slot-boundp-using-class
		slot-makunbound-using-class
		slot-definition-allocation
		slot-definition-initargs
		compute-slots)

  )

(in-package "ELE")

#+cmu
(eval-when (:compile-toplevel)
  (proclaim '(optimize (ext:inhibit-warnings 3))))