/*
;;;
;;; libsleepycat.c -- C wrappers for Sleepycat for FFI
;;; 
;;; Initial version 8/26/2004 by Ben Lee
;;; <blee@common-lisp.net>
;;; 
;;; part of
;;;
;;; Elephant: an object-oriented database for Common Lisp
;;;
;;; Copyright (c) 2004 by Andrew Blumberg and Ben Lee
;;; <ablumberg@common-lisp.net> <blee@common-lisp.net>
;;;
;;; This program is released under the following license
;;; ("GPL").  For differenct licensing terms, contact the
;;; copyright holders.
;;;
;;; This program is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU General
;;; Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at
;;; your option) any later version.
;;;
;;; This program is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE. See the GNU General Public License
;;; for more details.
;;;
;;; The GNU General Public License can be found in the file
;;; LICENSE which should have been distributed with this
;;; code.  It can also be found at
;;;
;;; http://www.opensource.org/licenses/gpl-license.php
;;;
;;; You should have received a copy of the GNU General
;;; Public License along with this program; if not, write
;;; to the Free Software Foundation, Inc., 59 Temple Place,
;;; Suite 330, Boston, MA 02111-1307 USA
;;;
;;; Portions of this program (namely the C unicode string
;;; sorter) are derived from IBM's ICU:
;;;
;;; http://oss.software.ibm.com/icu/
;;;
;;; Copyright (c) 1995-2003 International Business Machines
;;; Corporation and others All rights reserved.
;;;
;;; ICU's copyright, license and warranty can be found at
;;;
;;; http://oss.software.ibm.com/cvs/icu/~checkout~/icu/license.html
;;;
;;; or in the file LICENSE.
;;;
*/

/* Pointer arithmetic utility functions */
/* should these be in network-byte order? probably not..... */
int read_int(char *buf, int offset) {
  int i;
  memcpy(&i, buf+offset, sizeof(int));
  return i;
}

unsigned int read_uint(char *buf, int offset) {
  unsigned int ui; 
  memcpy(&ui, buf+offset, sizeof(unsigned int));
  return ui;
}

float read_float(char *buf, int offset) {
  float f;
  memcpy(&f, buf+offset, sizeof(float));
  return f;
}

double read_double(char *buf, int offset) {
  double d;
  memcpy(&d, buf+offset, sizeof(double));
  return d;
}

void write_int(char *buf, int num, int offset) {
  memcpy(buf+offset, &num, sizeof(int));
}

void write_uint(char *buf, unsigned int num, int offset) {
  memcpy(buf+offset, &num, sizeof(unsigned int));
}

void write_float(char *buf, float num, int offset) {
  memcpy(buf+offset, &num, sizeof(float));
}

void write_double(char *buf, double num, int offset) {
  memcpy(buf+offset, &num, sizeof(double));
}

char *offset_charp(char *p, int offset) {
  return p + offset;
}

#include <string.h>

void copy_buf(char *dest, int dest_offset, char *src, int src_offset, 
	      int length) {
  memcpy(dest + dest_offset, src + src_offset, length);
}


/* Sleepycat stuff */

#include <db.h>

/* Environment */

/* All "creation" functions return the create object, not the errno.
   This simplifies FFI handling. */

/* These next two functions are also needed because in db42 these
   are #define macros */
DB_ENV *db_env_cr(u_int32_t flags, int *errno) {
  DB_ENV *envp;
  *errno = db_env_create(&envp, flags);
  return envp; 
}

char * db_strerr(int error) {
  return db_strerror(error);
}

int db_env_close(DB_ENV *env, u_int32_t flags) {
  return env->close(env, flags);
}

int db_env_open(DB_ENV *env, char *home, u_int32_t flags, int mode) {
  return env->open(env, home, flags, mode);
}

int db_env_dbremove(DB_ENV *env, DB_TXN *txnid, char *file, char *database,
		    u_int32_t flags) {
  return env->dbremove(env, txnid, file, database, flags);
}

int db_env_dbrename(DB_ENV *env, DB_TXN *txnid, char *file, char *database,
		    char *newname, u_int32_t flags) {
  return env->dbrename(env, txnid, file, database, newname, flags);
}

int db_env_remove(DB_ENV *env, char *home, u_int32_t flags) {
  return env->remove(env, home, flags);
}

int db_env_set_flags(DB_ENV *dbenv, u_int32_t flags, int onoff) {
  return dbenv->set_flags(dbenv, flags, onoff);
}

int db_env_get_flags(DB_ENV *dbenv, u_int32_t *flagsp) {
  return dbenv->get_flags(dbenv, flagsp);
}

/* Database */

DB *db_cr(DB_ENV *dbenv, u_int32_t flags, int *errno) {
  DB *dbp;
  *errno = db_create(&dbp, dbenv, flags);
  return dbp;
}

int db_close(DB *db, u_int32_t flags) {
  return db->close(db, flags);
}

int db_open(DB *db, DB_TXN *txnid, char *file, char *database, DBTYPE type,
	    u_int32_t flags, int mode) {
  return db->open(db, txnid, file, database, type, flags, mode);
}

int db_remove(DB *db, char *file, char *database, u_int32_t flags) {
  return db->remove(db, file, database, flags);
}

int db_rename(DB *db, char *file, char *database, char *newname,
	      u_int32_t flags) {
  return db->rename(db, file, database, newname, flags);
}

int db_sync(DB *db, u_int32_t flags) {
  return db->sync(db, flags);
}

int db_truncate(DB *db, DB_TXN *txnid, u_int32_t *countp, u_int32_t flags) {
  return db->truncate(db, txnid, countp, flags);
}

int db_set_flags(DB *db, u_int32_t flags) {
  return db->set_flags(db, flags);
}

int db_get_flags(DB *db, u_int32_t *flagsp) {
  return db->get_flags(db, flagsp);
}

int db_set_pagesize(DB *db, u_int32_t pagesize) {
  return db->set_pagesize(db, pagesize);
}

int db_get_pagesize(DB *db, u_int32_t *pagesizep) {
  return db->get_pagesize(db, pagesizep);
}

int db_set_bt_compare(DB *db,
		      int (*bt_compare_fcn)(DB *db, const DBT *dbt1, 
					    const DBT *dbt2)) {
  return db->set_bt_compare(db, bt_compare_fcn);
}

int db_set_dup_compare(DB *db,
		       int (*dup_compare_fcn)(DB *db, const DBT *dbt1, 
					      const DBT *dbt2)) {
  return db->set_dup_compare(db, dup_compare_fcn);
}

#define type_numeric(c) ((c)<8)
#include <math.h>

double read_num(char *buf);

/* Inspired by the Sleepycat docs.  We have to memcpy to
   insure memory alignment. */
int lisp_compare(DB *dbp, const DBT *a, const DBT *b) {
  int difference;
  double ddifference;
  char *ad, *bd, at, bt;
  ad = (char*)a->data;
  bd = (char*)b->data;

  /* Compare OIDs. */
  difference = read_int(ad, 0) - read_int(bd, 0);
  if (difference) return difference;
  
  /* Have a type tag? */
  if (a->size == 4) 
    if (b->size == 4) 
      return 0;
    else
      return -1;
  else if (b->size == 4) 
    return 1;

  at = ad[4]; bt = bd[4];

  /* Compare numerics. */
  if (type_numeric(at) && type_numeric(bt)) {
    ddifference = read_num(ad+4) - read_num(bd+4);
    if (ddifference > 0) return 1;
    else if (ddifference < 0) return -1;
    return 0;
  }

  /* Compare types. */
  difference = at - bt;
  if (difference) return difference;

  /* Same type! */
  switch (at) {
  case 8: /* nil */
    return 0;
  case 9: /* 8-bit symbol */
  case 10: /* 8-bit string */
  case 11: /* 8-bit pathname */
    return case_cmp(ad+9, read_int(ad, 5), bd+9, read_int(bd, 5));
  case 12: /* 16-bit symbol */
  case 13: /* 16-bit string */
  case 14: /* 16-bit pathname */
    return utf16_cmp(ad+9, read_int(ad, 5), bd+9, read_int(bd, 5));
  default:
    return lex_cmp(ad+5, (a->size)-5, bd+5, (b->size)-5);
  }
}

int db_set_lisp_compare(DB *db) {
  return db->set_bt_compare(db, &lisp_compare);
}

int db_set_lisp_dup_compare(DB *db) {
  return db->set_dup_compare(db, &lisp_compare);
}

#ifndef exp2
#define exp2(c) (pow(2,(c)))
#endif

double read_num(char *buf) {
  char *limit;
  double i, result, denom;
  switch (buf[0]) {
  case 1:
  case 2:
    return (double)read_int(buf, 1);
  case 3:
    return (double)read_float(buf, 1);
  case 4:
    return read_double(buf, 1);
  case 5:
    result = 0;
    buf += 5;
    limit = buf + read_uint(buf, -4);
    for(i=0 ; buf < limit; i++, buf = buf+4) {
      result -= exp2(i*32) * read_uint(buf, 0);
    }
    return result;
  case 6:
    result = 0;
    buf += 5;
    limit = buf + read_uint(buf, -4);
    for(i=0 ; buf < limit; i++, buf = buf+4) {
      result += exp2(i*32) * read_uint(buf, 0);
    }
    return result;
  case 7:
    switch ((++buf)[0]) {
    case 1: 
      result = (double)read_int(++buf, 0);
      buf += 4;
      break;
    case 5:
      result = 0;
      buf += 5;
      limit = buf + read_uint(buf, -4);
      for(i=0 ; buf < limit; i++, buf = buf+4) {
	result -= exp2(i*32) - read_uint(buf, 0);
      }
      break;
    case 6:
      result = 0;
      buf += 5;
      limit = buf + read_uint(buf, -4);
      for(i=0 ; buf < limit; i++, buf = buf+4) {
	result += exp2(i*32) * read_uint(buf, 0);
      }
      break;
    }
    
    switch (buf[0]) {
    case 1: 
      return result / read_int(++buf, 0);
    case 5:
      denom = 0;
      buf += 5;
      limit = buf + read_uint(buf, -4);
      for(i=0 ; buf < limit; i++, buf = buf+4) {
	denom -= exp2(i*32) * read_uint(buf, 0);
      }
      return result / denom;
    case 6:
      denom = 0;
      buf += 5;
      limit = buf + read_uint(buf, -4);
      for(i=0 ; buf < limit; i++, buf = buf+4) {
	denom += exp2(i*32) * read_uint(buf, 0);
      }
      return result / denom;
    }    
  }
}

int case_cmp(const char *a, int32_t length1, const char *b, int32_t length2) {
  int min, sizediff, diff;
  sizediff = length1 - length2;
  min = sizediff > 0 ? length2 : length1;
  diff = strncasecmp(a, b, min);
  if (diff == 0) return sizediff;
  return diff;
}

int lex_cmp(const char *a, int32_t length1, const char *b, int32_t length2) {
  int min, sizediff, diff;
  sizediff = length1 - length2;
  min = sizediff > 0 ? length2 : length1;
  diff = memcmp(a, b, min);
  if (diff == 0) return sizediff;
  return diff;
}

/* The following is derived from
   http://oss.software.ibm.com/cvs/icu/~checkout~/icu/source/common/ustring.c 
*/
typedef uint16_t UChar;

#define UTF_IS_LEAD(c) (((c)&0xfffffc00)==0xd800)
#define UTF_IS_TRAIL(c) (((c)&0xfffffc00)==0xdc00)

/* compare UTF-16 strings */
/* memcmp/UnicodeString style, both length-specified */
/* don't assume byte-aligned! */
int utf16_cmp(const char *s1, int32_t length1, 
	      const char *s2, int32_t length2) {
  const char *start1, *start2, *limit1, *limit2;
  UChar c1, c2;
  int32_t lengthResult;

  if(length1<length2) {
    lengthResult=-1;
    limit1=s1+length1;
  } else if(length1==length2) {
    lengthResult=0;
    limit1=s1+length1;
  } else /* length1>length2 */ {
    lengthResult=1;
    limit1=s1+length2;
  }

  if(s1==s2) return lengthResult;

  start1=s1;
  start2=s2;

  for(;;) {
    if(s1==limit1) return lengthResult;

    memcpy(&c1, s1, sizeof(UChar));
    memcpy(&c2, s2, sizeof(UChar));
    if(c1!=c2) break;

    s1 = s1 + 2;
    s2 = s2 + 2;
  }

  limit1=start1+length1;
  limit2=start2+length2;

  if(c1>=0xd800 && c2>=0xd800) {
    if(c1>=0xe000) 
      c1-=0x800;
    else
      c1+=0x2000;
    
    if(c2>=0xe000)
      c2-=0x800;
    else
      c2+=0x2000;
      
    /* here's some newer code which i can't make work
    if((c1<=0xdbff && (s1+1)!=limit1 && UTF_IS_TRAIL(*(s1+1))) ||
       (UTF_IS_TRAIL(c1) && start1!=s1 && UTF_IS_LEAD(*(s1-1)))) {
    } else {
      c1-=0x2800;
    }

    if((c2<=0xdbff && (s2+1)!=limit2 && UTF_IS_TRAIL(*(s2+1))) ||
       (UTF_IS_TRAIL(c2) && start2!=s2 && UTF_IS_LEAD(*(s2-1)))) {
    } else {
      c2-=0x2800;
      }*/
  }

  return (int32_t)c1-(int32_t)c2;
}


/* Accessors */
/* We manage our own buffers (DB_DBT_USERMEM). */

int db_get_raw(DB *db, DB_TXN *txnid, 
	       char *key, u_int32_t key_size,
	       char *buffer, u_int32_t buffer_length,
	       u_int32_t flags, u_int32_t *result_size) {
  DBT DBTKey, DBTValue;
  int ret;
  
  memset(&DBTKey, 0, sizeof(DBT));
  memset(&DBTValue, 0, sizeof(DBT));
  DBTKey.data = key;
  DBTKey.size = key_size;
  DBTValue.data = buffer;
  DBTValue.ulen = buffer_length;
  DBTValue.flags |= DB_DBT_USERMEM;
  
  ret = db->get(db, txnid, &DBTKey, &DBTValue, flags);
  *result_size = DBTValue.size;
  
  return ret;
}

int db_put_raw(DB *db, DB_TXN *txnid, 
	       char *key, u_int32_t key_size,
	       char *value, u_int32_t value_size,
	       u_int32_t flags) {
  DBT DBTKey, DBTValue;
  
  memset(&DBTKey, 0, sizeof(DBT));
  memset(&DBTValue, 0, sizeof(DBT));
  DBTKey.data = key;
  DBTKey.size = key_size;
  DBTValue.data = value;
  DBTValue.size = value_size;
  
  return db->put(db, txnid, &DBTKey, &DBTValue, flags);
}

int db_del(DB *db, DB_TXN *txnid, 
	   char *key, u_int32_t key_size,
	   u_int32_t flags) {
  DBT DBTKey;
  
  memset(&DBTKey, 0, sizeof(DBT));
  DBTKey.data = key;
  DBTKey.size = key_size;
  return db->del(db, txnid, &DBTKey, flags);
}


/* Cursors */

DBC * db_cursor(DB *db, DB_TXN *txnid, u_int32_t flags, int *errno) {
  DBC *cursor;
  *errno = db->cursor(db, txnid, &cursor, flags);
  return cursor;
}

int db_cursor_close(DBC *cursor) {
  return cursor->c_close(cursor);
}

int db_cursor_del(DBC *cursor, u_int32_t flags) {
  return cursor->c_del(cursor, flags);
}

DBC * db_cursor_dup(DBC *cursor, u_int32_t flags, int *errno) {
  DBC *dup;
  *errno = cursor->c_dup(cursor, &dup, flags);
  return dup;
}

int db_cursor_get_raw(DBC *cursor, 
		      char *keybuf, u_int32_t keybuf_size,
		      u_int32_t keybuf_length,
		      char *buffer, u_int32_t buffer_size,
		      u_int32_t buffer_length,
		      u_int32_t flags, u_int32_t *ret_key_size,
		      u_int32_t *result_size) {
  DBT DBTKey, DBTValue;
  int ret;
  
  memset(&DBTKey, 0, sizeof(DBT));
  memset(&DBTValue, 0, sizeof(DBT));
  DBTKey.data = keybuf;
  DBTKey.size = keybuf_size;
  DBTKey.ulen = keybuf_length;
  DBTKey.flags |= DB_DBT_USERMEM;
  DBTValue.data = buffer;
  DBTValue.size = buffer_size;
  DBTValue.ulen = buffer_length;
  DBTValue.flags |= DB_DBT_USERMEM;
  
  ret = cursor->c_get(cursor, &DBTKey, &DBTValue, flags);
  *ret_key_size = DBTKey.size;
  *result_size = DBTValue.size;
  
  return ret;
}

int db_cursor_pget_raw(DBC *cursor, 
		       char *keybuf, u_int32_t keybuf_size,
		       u_int32_t keybuf_length,
		       char *pkeybuf, u_int32_t pkeybuf_size,
		       u_int32_t pkeybuf_length,
		       char *buffer, u_int32_t buffer_size,
		       u_int32_t buffer_length,
		       u_int32_t flags, 
		       u_int32_t *ret_key_size,
		       u_int32_t *ret_pkey_size,
		       u_int32_t *result_size) {
  DBT DBTKey, DBTPKey, DBTValue;
  int ret;
  
  memset(&DBTKey, 0, sizeof(DBT));
  memset(&DBTPKey, 0, sizeof(DBT));
  memset(&DBTValue, 0, sizeof(DBT));
  DBTKey.data = keybuf;
  DBTKey.size = keybuf_size;
  DBTKey.ulen = keybuf_length;
  DBTKey.flags |= DB_DBT_USERMEM;
  DBTPKey.data = pkeybuf;
  DBTPKey.size = pkeybuf_size;
  DBTPKey.ulen = pkeybuf_length;
  DBTPKey.flags |= DB_DBT_USERMEM;
  DBTValue.data = buffer;
  DBTValue.size = buffer_size;
  DBTValue.ulen = buffer_length;
  DBTValue.flags |= DB_DBT_USERMEM;
  
  ret = cursor->c_pget(cursor, &DBTKey, &DBTPKey, &DBTValue, flags);
  *ret_key_size = DBTKey.size;
  *ret_pkey_size = DBTPKey.size;
  *result_size = DBTValue.size;
  
  return ret;
}

int db_cursor_put_raw(DBC *cursor,
		      char *key, u_int32_t key_size,
		      char *value, u_int32_t value_size,
		      u_int32_t flags) {
  DBT DBTKey, DBTValue;
  
  memset(&DBTKey, 0, sizeof(DBT));
  memset(&DBTValue, 0, sizeof(DBT));
  DBTKey.data = key;
  DBTKey.size = key_size;
  DBTValue.data = value;
  DBTValue.size = value_size;
  
  return cursor->c_put(cursor, &DBTKey, &DBTValue, flags);
}


/* Silently does nothing if the key/value isn't found.
   Can't use auto-commit here! */
int db_del_kv(DB *db, DB_TXN *tid, 
	      char *key, u_int32_t key_size,
	      char *value, u_int32_t value_size) {
  DBT DBTKey, DBTValue;
  DBC *cursor;
  int ret, c_ret;

  memset(&DBTKey, 0, sizeof(DBT));
  DBTKey.data = key;
  DBTKey.size = key_size;
  memset(&DBTValue, 0, sizeof(DBT));
  DBTValue.data = value;
  DBTValue.size = value_size;
  
  if ((ret = db->cursor(db, tid, &cursor, 0)) != 0)
    return ret;

  if ((ret = cursor->c_get(cursor, &DBTKey, &DBTValue, DB_GET_BOTH)) != 0)
    goto fail;

  ret = cursor->c_del(cursor, 0);

 fail:
  if ((c_ret = cursor->c_close(cursor)) != 0)
    return c_ret;
  return ret;
}

/* Bulk retrieval */

int db_cursor_get_multiple_key(DBC *cursor, 
			       char *keybuf, u_int32_t keybuf_size,
			       u_int32_t keybuf_length,
			       char *buffer, u_int32_t buffer_size,
			       u_int32_t buffer_length,
			       u_int32_t flags, u_int32_t *ret_key_size,
			       u_int32_t *result_size,
			       void **pointer, DBT **data) {
  DBT DBTKey, DBTValue;
  int ret;
  
  memset(&DBTKey, 0, sizeof(DBT));
  memset(&DBTValue, 0, sizeof(DBT));
  DBTKey.data = keybuf;
  DBTKey.size = keybuf_size;
  DBTKey.ulen = keybuf_length;
  DBTKey.flags |= DB_DBT_USERMEM;
  DBTValue.data = buffer;
  DBTValue.size = buffer_size;
  DBTValue.ulen = buffer_length;
  DBTValue.flags |= DB_DBT_USERMEM;
  
  flags |= DB_MULTIPLE_KEY;
  ret = cursor->c_get(cursor, &DBTKey, &DBTValue, flags);
  *ret_key_size = DBTKey.size;
  *result_size = DBTValue.size;
  if ((DBTKey.size <= DBTKey.ulen) && (DBTValue.size <= DBTValue.ulen)) {
    **data = DBTValue;
    DB_MULTIPLE_INIT(*pointer, *data);
  }
				    
  return ret;
}

void db_multiple_key_next(void *pointer, DBT *data,
			  char **key, u_int32_t *ret_key_size,
			  char **result, u_int32_t *result_size) {
  DB_MULTIPLE_KEY_NEXT(pointer, data, 
		       *key, *ret_key_size, 
		       *result, *result_size);
}

/* Transactions */

DB_TXN * db_txn_begin(DB_ENV *env, DB_TXN *parent, 
		      u_int32_t flags, int *errno) {
  DB_TXN * p;
  *errno = env->txn_begin(env, parent, &p, flags);
  return p;
}

int db_txn_abort(DB_TXN *txnid) {
  return txnid->abort(txnid);
}

int db_txn_commit(DB_TXN *txnid, u_int32_t flags) {
  return txnid->commit(txnid, flags);
}


int db_txnp_begin(DB_ENV *env, DB_TXN *parent, DB_TXN **txnp,
		 u_int32_t flags) {
  return env->txn_begin(env, parent, txnp, flags);
}

/* Locks and timeouts */

u_int32_t db_txn_id(DB_TXN *tid) {
  return tid->id(tid);
}

int db_env_lock_id(DB_ENV *env, u_int32_t *idp) {
  return env->lock_id(env, idp);
}

int db_env_lock_id_free(DB_ENV *env, u_int32_t id) {
  return env->lock_id_free(env, id);
}

int db_env_lock_get(DB_ENV *env, u_int32_t locker,
		    u_int32_t flags, char *object, u_int32_t object_size,
		    const db_lockmode_t lock_mode, DB_LOCK *lock) {
  DBT DBTObject;
  memset(&DBTObject, 0, sizeof(DBT));
  DBTObject.data = object;
  DBTObject.size = object_size;

  return env->lock_get(env, locker, flags, &DBTObject, lock_mode, lock);
}

int db_env_lock_put(DB_ENV *env, DB_LOCK *lock) {
  return env->lock_put(env, lock);
}

int db_env_lock_vec(DB_ENV *env, u_int32_t locker, u_int32_t flags,
		    DB_LOCKREQ list[], int nlist, DB_LOCKREQ **elistp) {
  return env->lock_vec(env, locker, flags, list, nlist, elistp);
}

/* db_timeout_t = u_int32_t */
int db_env_set_timeout(DB_ENV *env, db_timeout_t timeout, u_int32_t flags) {
  return env->set_timeout(env, timeout, flags);
}

int db_env_get_timeout(DB_ENV *env, db_timeout_t *timeoutp, u_int32_t flags) {
  return env->get_timeout(env, timeoutp, flags);
}

int db_env_set_lk_detect(DB_ENV *env, u_int32_t detect) {
    return env->set_lk_detect(env, detect);
}

int db_env_get_lk_detect(DB_ENV *env, u_int32_t *detectp) {
    return env->get_lk_detect(env, detectp);
}

int db_env_lock_detect(DB_ENV *env, u_int32_t flags, u_int32_t atype,
		   int *aborted) {
    return env->lock_detect(env, flags, atype, aborted);
}

/* Secondary indices */

int db_associate(DB *primary, DB_TXN *txnid, DB *secondary,
		 int (*callback)(DB *, const DBT *, const DBT *, DBT *),
		 u_int32_t flags) {
  return primary->associate(primary, txnid, secondary, callback, flags);
}

int never_index(DB *db, const DBT *key, const DBT *data, DBT *result) {
  return DB_DONOTINDEX;
}

int db_fake_associate(DB *primary, DB_TXN *txnid, DB *secondary, 
		      u_int32_t flags) {
  return primary->associate(primary, txnid, secondary, &never_index, flags);
}

/* Poor man's counters */

int next_counter(DB_ENV *env, DB *db, DB_TXN *parent,
		 char *key, u_int32_t key_size,
		 char *lockid, u_int32_t lockid_size) {
  DB_LOCK lock;
  DBT DBTKey, DBTData; 
  DB_TXN *tid; 
  int counter, tries, ret, t_ret, lockheld;
  u_int32_t id;

  /* Initialization. */ 
  memset(&lock, 0, sizeof(lock)); 
  memset(&DBTKey, 0, sizeof(DBTKey)); 
  memset(&DBTData, 0, sizeof(DBTData)); 
  DBTKey.data = key;
  DBTKey.size = key_size;
  DBTData.data = lockid;
  DBTData.size = lockid_size;

  tries = 0;

 loop:
  lockheld = 0;

  /* Begin the transaction. */ 
  if ((ret = env->txn_begin(env, parent, &tid, 0)) != 0) { 
    env->err(env, ret, "DB_ENV->txn_begin"); 
    return (-1); 
  }

  id = tid->id(tid);

  if ((ret = env->lock_get(env, id, 0, &DBTData, DB_LOCK_WRITE, &lock)) != 0)
    goto fail;

  lockheld = 1;

  memset(&DBTData, 0, sizeof(DBTData)); 
  DBTData.data = &counter;
  DBTData.ulen = sizeof(counter);
  DBTData.flags |= DB_DBT_USERMEM;

  if ((ret = db->get(db, tid, &DBTKey, &DBTData, 0)) != 0) 
    goto fail;

  ++counter;

  memset(&DBTData, 0, sizeof(DBTData)); 
  DBTData.data = &counter;
  DBTData.size = sizeof(counter);

  if ((ret = db->put(db, tid, &DBTKey, &DBTData, 0)) != 0) 
    goto fail;

  if ((ret = env->lock_put(env, &lock)) != 0) 
    goto fail;

  if ((ret = tid->commit(tid, DB_TXN_NOSYNC)) != 0) { 
    env->err(env, ret, "DB_TXN->commit"); 
    return (-2); 
  } 
  return (counter); 


 fail:
  if (lockheld)
    if ((ret = env->lock_put(env, &lock)) != 0) 
      return (-3);

  /* Abort and retry the operation. */ 
  if ((t_ret = tid->abort(tid)) != 0) { 
    env->err(env, t_ret, "DB_TXN->abort"); 
    return (-4);
  } 
  if (tries++ == 100)
    return (-5); 
  goto loop;
}
