;;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp; Base: 10 -*-
;;;
;;; sleepycat.lisp -- FFI interface to Berkeley DB
;;; 
;;; Initial version 8/26/2004 by Ben Lee
;;; <blee@common-lisp.net>
;;; 
;;; part of
;;;
;;; Elephant: an object-oriented database for Common Lisp
;;;
;;; Copyright (c) 2004 by Andrew Blumberg and Ben Lee
;;; <ablumberg@common-lisp.net> <blee@common-lisp.net>
;;;
;;; This program is released under the following license
;;; ("GPL").  For differenct licensing terms, contact the
;;; copyright holders.
;;;
;;; This program is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU General
;;; Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at
;;; your option) any later version.
;;;
;;; This program is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE. See the GNU General Public License
;;; for more details.
;;;
;;; The GNU General Public License can be found in the file
;;; LICENSE which should have been distributed with this
;;; code.  It can also be found at
;;;
;;; http://www.opensource.org/licenses/gpl-license.php
;;;
;;; You should have received a copy of the GNU General
;;; Public License along with this program; if not, write
;;; to the Free Software Foundation, Inc., 59 Temple Place,
;;; Suite 330, Boston, MA 02111-1307 USA
;;;


(defpackage sleepycat
  (:documentation "A low-level UFFI-based interface to
Berkeley DB / Sleepycat, via the libsleepycat.c wrapper.
Partly intended to be usable outside Elephant, but with some
magic for Elephant.  In general there is a 1-1 mapping from
functions here and functions in Sleepycat, so refer to their
documentation for details.")
  (:use common-lisp uffi)
  #+cmu
  (:use alien)
  #+sbcl
  (:use sb-alien)
  #+cmu
  (:import-from :sys
		#:sap+)
  #+sbcl
  (:import-from :sb-sys
		#:sap+)  
  #+openmcl
  (:import-from :ccl
		#:byte-length)
  (:export #:*current-transaction* 
	   
	   #:buffer-stream #:make-buffer-stream #:with-buffer-streams
	   #:resize-buffer-stream #:resize-buffer-stream-no-copy 
	   #:reset-buffer-stream #:buffer-stream-buffer
	   #:buffer-write-byte #:buffer-write-int 
	   #:buffer-write-uint #:buffer-write-float #:buffer-write-double 
	   #:buffer-write-string #:buffer-read-byte #:buffer-read-fixnum 
	   #:buffer-read-int #:buffer-read-uint #:buffer-read-float 
	   #:buffer-read-double #:buffer-read-string #:byte-length
	   
	   #:pointer-int #:pointer-void #:array-or-pointer-char
	   
	   #:db-env-create #:db-env-close #:db-env-open #:db-env-dbremove
	   #:db-env-dbrename #:db-env-remove 
	   #:db-env-set-flags #:db-env-get-flags
	   #:db-create #:db-close #:db-open 
	   #:db-remove #:db-rename #:db-sync #:db-truncate
	   #:db-set-flags #:db-get-flags
	   #:db-get-key-buffered #:db-get-buffered #:db-get 
	   #:db-put-buffered #:db-put 
	   #:db-delete-buffered #:db-delete 
	   #:db-cursor #:db-cursor-close #:db-cursor-delete
	   #:db-cursor-duplicate 
	   #:db-cursor-move-buffered #:db-cursor-set-buffered
	   #:db-cursor-get-both-buffered 
	   #:db-cursor-pmove-buffered #:db-cursor-pset-buffered
	   #:db-cursor-pget-both-buffered #:db-cursor-put-buffered
	   #:db-transaction-begin #:db-transaction-abort 
	   #:db-transaction-commit #:with-transaction 
	   #:db-transaction-id #:db-env-lock-id #:db-env-lock-id-free
	   #:db-env-lock-get #:db-env-lock-put #:with-lock
	   #:db-env-set-timeout #:db-env-get-timeout
	   #:db-env-set-lock-detect #:db-env-get-lock-detect
	   #:db-error #:db-error-errno
	   
	   #:+NULL-VOID+ #:+NULL-CHAR+
	   #:DB-BTREE #:DB-HASH #:DB-QUEUE #:DB-RECNO #:DB-UNKNOWN 
	   #:DB_KEYEMPTY #:DB_LOCK_DEADLOCK #:DB_LOCK_NOTGRANTED 
	   #:DB_NOTFOUND
	   #:DB-LOCKMODE#NG #:DB-LOCKMODE#READ #:DB-LOCKMODE#WRITE 
	   #:DB-LOCKMODE#wAIT #:DB-LOCKMODE#IWRITE #:DB-LOCKMODE#IREAD
	   #:DB-LOCKMODE#IWR #:DB-LOCKMODE#DIRTY #:DB-LOCKMODE#WWRITE
	   ))

(in-package "SLEEPYCAT")

#+cmu
(eval-when (:compile-toplevel)
  (proclaim '(optimize (ext:inhibit-warnings 3))))

(eval-when (:compile-toplevel :load-toplevel)
  ;; UFFI
  ;;(asdf:operate 'asdf:load-op :uffi)

  ;; DSO loading - Edit these for your system!

  ;; Under linux you may need to load some kind of pthread
  ;; library.  I can't figure out which is the right one.
  ;; This one worked for me.  There are known issues with
  ;; Red Hat and Berkeley DB, search google.
  #+linux
  (unless 
      (uffi:load-foreign-library "/lib/tls/libpthread.so.0" :module "pthread")
    (error "Couldn't load libpthread!"))

  (unless
      (uffi:load-foreign-library 
       ;; Sleepycat: this works on linux
       #+linux
       "/usr/local/BerkeleyDB.4.2/lib/libdb.so" 
       ;; this works on FreeBSD
       #+(and (or bsd freebsd) (not darwin))
       "/usr/local/lib/db42/libdb.so" 
       #+darwin
       "/usr/local/BerkeleyDB.4.2/lib/libdb.dylib" 
       :module "sleepycat")
    (error "Couldn't load libdb (Sleepycat)!"))

  ;; Libsleepycat.so: edit this
  (unless
      (uffi:load-foreign-library 
       (if (find-package 'asdf)
	   (merge-pathnames 
	    #p"libsleepycat.so"
	    (asdf:component-pathname (asdf:find-system 'elephant)))
	   "/usr/local/share/common-lisp/elephant-0.2/libsleepycat.so")
       :module "libsleepycat")
    (error "Couldn't load libsleepycat!"))

  ;; fini on user editable part

  (def-type pointer-int (* :int))
  (def-type pointer-void :pointer-void)
  (def-foreign-type array-or-pointer-char
      #+allegro (:array :char)
      #+(or cmu sbcl scl openmcl) (* :char))
  (def-type array-or-pointer-char array-or-pointer-char)
  (def-enum DBTYPE ((:BTREE 1) :HASH :QUEUE :RECNO :UNKNOWN))
)

(declaim (inline read-int read-uint read-float read-double 
		 write-int write-uint write-float write-double
		 offset-char-pointer copy-str-to-buf copy-bufs
		 ;;resize-buffer-stream 
		 ;;buffer-stream-buffer buffer-stream-size buffer-stream-position
		 ;;buffer-stream-length 
		 reset-buffer-stream
		 buffer-write-byte buffer-write-int buffer-write-uint
		 buffer-write-float buffer-write-double buffer-write-string
		 buffer-read-byte buffer-read-fixnum buffer-read-int
		 buffer-read-uint buffer-read-float buffer-read-double 
		 buffer-read-string))

;; Constants and Flags
;; eventually write a macro which generates a custom flag function.

;I don't like the UFFI syntax for enumerations
(defconstant DB-BTREE                 1)
(defconstant DB-HASH                  2)
(defconstant DB-QUEUE                 3)
(defconstant DB-RECNO                 4)
(defconstant DB-UNKNOWN               5)

(defconstant DB_AUTO_COMMIT   #x1000000)
(defconstant DB_JOINENV	      #x0040000)
(defconstant DB_INIT_CDB      #x0001000)
(defconstant DB_INIT_LOCK     #x0002000)
(defconstant DB_INIT_LOG      #x0004000)
(defconstant DB_INIT_MPOOL    #x0008000)
(defconstant DB_INIT_REP      #x0010000)
(defconstant DB_INIT_TXN      #x0020000)
(defconstant DB_RECOVER	      #x0000020)
(defconstant DB_RECOVER_FATAL #x0200000)
(defconstant DB_LOCKDOWN      #x0080000)
(defconstant DB_PRIVATE	      #x0100000)
(defconstant DB_SYSTEM_MEM    #x0400000)
(defconstant DB_THREAD	      #x0000040)
(defconstant DB_FORCE	      #x0000004)
(defconstant DB_DIRTY_READ    #x2000000)
(defconstant DB_CREATE	      #x0000001)
(defconstant DB_EXCL          #x0001000)
(defconstant DB_NOMMAP	      #x0000008)
(defconstant DB_RDONLY	      #x0000010)
(defconstant DB_TRUNCATE      #x0000080)
(defconstant DB_TXN_NOSYNC    #x0000100)
(defconstant DB_TXN_NOWAIT    #x0001000)
(defconstant DB_TXN_SYNC      #x0002000)
(defconstant DB_LOCK_NOWAIT   #x001)
(defconstant DB_DUP	      #x0000002)
(defconstant DB_DUPSORT	      #x0000004)

(defconstant DB_CURRENT		      7)
(defconstant DB_FIRST		      9)
(defconstant DB_GET_BOTH	     10)
(defconstant DB_GET_BOTH_RANGE	     12)
(defconstant DB_LAST		     17)
(defconstant DB_NEXT		     18)
(defconstant DB_NEXT_DUP	     19)
(defconstant DB_NEXT_NODUP	     20)
(defconstant DB_PREV		     25)
(defconstant DB_PREV_NODUP	     26)
(defconstant DB_SET		     28)
(defconstant DB_SET_RANGE	     30)

(defconstant DB_AFTER		      1)
(defconstant DB_BEFORE		      3)
(defconstant DB_KEYFIRST	     15)
(defconstant DB_KEYLAST		     16)

(defconstant DB_NODUPDATA	     21)
(defconstant DB_NOOVERWRITE	     22)
(defconstant DB_NOSYNC		     23)

(defconstant DB_POSITION	     24)

(defconstant DB_SET_LOCK_TIMEOUT     29)
(defconstant DB_SET_TXN_TIMEOUT      33)

(defconstant DB_KEYEMPTY         -30997)
(defconstant DB_KEYEXIST	 -30996)
(defconstant DB_LOCK_DEADLOCK    -30995)
(defconstant DB_LOCK_NOTGRANTED  -30994)
(defconstant DB_NOTFOUND         -30990)

(defconstant DB_LOCK_DEFAULT	     1)
(defconstant DB_LOCK_EXPIRE	     2)
(defconstant DB_LOCK_MAXLOCKS        3)
(defconstant DB_LOCK_MINLOCKS        4)
(defconstant DB_LOCK_MINWRITE        5)
(defconstant DB_LOCK_OLDEST	     6)
(defconstant DB_LOCK_RANDOM	     7)
(defconstant DB_LOCK_YOUNGEST        8)

(defvar +NULL-VOID+ (make-null-pointer :void)
  "A null pointer to a void type.")
(defvar +NULL-CHAR+ (make-null-pointer :char)
  "A null pointer to a char type.")


(def-enum DB-LOCKOP ((:DUMP 0) :GET :GET-TIMEOUT :INHERIT 
		     :PUT :PUT-ALL :PUT-OBJ :PUT-READ
		     :TIMEOUT :TRADE :UPGRADE-WRITE))

(def-enum DB-LOCKMODE ((:NG 0) :READ :WRITE :WAIT 
		       :IWRITE :IREAD :IWR :DIRTY :WWRITE))

(def-struct DB-LOCK
    (off :unsigned-int)
  (ndx :unsigned-int)
  (gen :unsigned-int)
  (mode DB-LOCKMODE))

#+openmcl
(ccl:def-foreign-type DB-LOCK (:struct DB-LOCK))

(def-struct DB-LOCKREQ
    (op DB-LOCKOP)
  (mode DB-LOCKMODE)
  (timeout :unsigned-int)
  (obj (:array :char))
  (lock (* DB-LOCK)))

#+openmcl
(ccl:def-foreign-type DB-LOCKREQ (:struct DB-LOCKREQ))


;; Thread local storage (special variables)

(defvar *current-transaction* +NULL-VOID+
  "The transaction which is currently in effect.")

(defvar *errno-buffer* (allocate-foreign-object :int 1)
  "Resourced space for errno return values.")
(defvar *buffer-streams* (make-array 0 :adjustable t :fill-pointer t)
  "Vector of buffer-streams, which you can grab / return.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; buffer-streams
;;;
;;; a stream-like interface for our buffers; methods are
;;; below.  ultimately we might want a gray / simple -stream
;;; for real, for now who cares?

(defstruct buffer-stream
  "A stream-like interface to foreign (alien) char buffers."
  (buffer (allocate-foreign-object :char 10) :type array-or-pointer-char)
  (size 0 :type fixnum)
  (position 0 :type fixnum)
  (length 10 :type fixnum))

(defun grab-buffer-stream ()
  "Grab a buffer-stream from the *buffer-streams* resource pool."
  (declare (optimize (speed 3)))
  (if (= (length *buffer-streams*) 0)
      (make-buffer-stream)
      (vector-pop *buffer-streams*)))

(defun return-buffer-stream (bs)
  "Return a buffer-stream to the *buffer-streams* resource pool."
  (declare (optimize (speed 3)))
  (reset-buffer-stream bs)
  (vector-push-extend bs *buffer-streams*))

(defmacro with-buffer-streams (names &body body)
  "Grab a buffer-stream, executes forms, and returns the
stream to the pool on exit."
  `(let ,(loop for name in names collect (list name '(grab-buffer-stream)))
    (unwind-protect
	 (progn ,@body)
      (progn
	,@(loop for name in names 
		collect (list 'return-buffer-stream name))))))

;; Buffer management / pointer arithmetic

;; Notes: on Allegro: with-cast-pointer + deref-array is
;; faster than FFI + C pointer arithmetic.  however pointer
;; arithmetic is usually consing.  OpenMCL supports
;; non-consing pointer arithmentic though.  Check these
;; CMUCL / SBCL things don't cons unless necessary.

;; TODO: #+openmcl versions which do macptr arith.  

#+(or cmu sbcl)
(defun read-int (buf offset)
  "Read a 32-bit signed integer from a foreign char buffer."
  (declare (optimize (speed 3) (safety 0))
	   (type (alien (* char)) buf)
	   (type fixnum offset))
  (the (signed-byte 32)
    (deref (cast (sap-alien (sap+ (alien-sap buf) offset) (* char))
		 (* integer)))))

#+(or cmu sbcl)
(defun read-uint (buf offset)
  "Read a 32-bit unsigned integer from a foreign char buffer."
  (declare (optimize (speed 3) (safety 0))
	   (type (alien (* char)) buf)
	   (type fixnum offset))
  (the (unsigned-byte 32)
    (deref (cast (sap-alien (sap+ (alien-sap buf) offset) (* char))
		 (* (unsigned 32))))))

#+(or cmu sbcl)
(defun read-float (buf offset)
  "Read a single-float from a foreign char buffer."
  (declare (optimize (speed 3) (safety 0))
	   (type (alien (* char)) buf)
	   (type fixnum offset))
  (the single-float
    (deref (cast (sap-alien (sap+ (alien-sap buf) offset) (* char))
		 (* single-float)))))

#+(or cmu sbcl)
(defun read-double (buf offset)
  "Read a double-float from a foreign char buffer."
  (declare (optimize (speed 3) (safety 0))
	   (type (alien (* char)) buf)
	   (type fixnum offset))
  (the double-float
    (deref (cast (sap-alien (sap+ (alien-sap buf) offset) (* char))
		 (* double-float)))))

#+(or cmu sbcl)
(defun write-int (buf num offset)
  "Write a 32-bit signed integer to a foreign char buffer."
  (declare (optimize (speed 3) (safety 0))
	   (type (alien (* char)) buf)
	   (type (signed-byte 32) num)
	   (type fixnum offset))
  (setf (deref (cast (sap-alien (sap+ (alien-sap buf) offset) (* char))
		     (* integer))) num))

#+(or cmu sbcl)
(defun write-uint (buf num offset)
  "Write a 32-bit unsigned integer to a foreign char buffer."
  (declare (optimize (speed 3) (safety 0))
	   (type (alien (* char)) buf)
	   (type (unsigned-byte 32) num)
	   (type fixnum offset))
  (setf (deref (cast (sap-alien (sap+ (alien-sap buf) offset) (* char))
		     (* (unsigned 32)))) num))

#+(or cmu sbcl)
(defun write-float (buf num offset)
  "Write a single-float to a foreign char buffer."
  (declare (optimize (speed 3) (safety 0))
	   (type (alien (* char)) buf)
	   (type single-float num)
	   (type fixnum offset))
  (setf (deref (cast (sap-alien (sap+ (alien-sap buf) offset) (* char))
		     (* single-float))) num))

#+(or cmu sbcl)
(defun write-double (buf num offset)
  "Write a double-float to a foreign char buffer."
  (declare (optimize (speed 3) (safety 0))
	   (type (alien (* char)) buf)
	   (type double-float num)
	   (type fixnum offset))
  (setf (deref (cast (sap-alien (sap+ (alien-sap buf) offset) (* char))
		     (* double-float))) num))

#+(or cmu sbcl)
(defun offset-char-pointer (p offset)
  "Pointer arithmetic."
  (declare (optimize (speed 3) (safety 0))
	   (type (alien (* char)) p)
	   (type fixnum offset))
  (sap-alien (sap+ (alien-sap p) offset) (* char)))

#-(or cmu sbcl)
(def-function ("read_int" read-int)
    ((buf array-or-pointer-char)
     (offset :int))
  :returning :int)

#-(or cmu sbcl)
(def-function ("read_uint" read-uint)
    ((buf array-or-pointer-char)
     (offset :int))
  :returning :unsigned-int)

#-(or cmu sbcl)
(def-function ("read_float" read-float)
    ((buf array-or-pointer-char)
     (offset :int))
  :returning :float)

#-(or cmu sbcl)
(def-function ("read_double" read-double)
    ((buf array-or-pointer-char)
     (offset :int))
  :returning :double)

#-(or cmu sbcl)
(def-function ("write_int" write-int)
    ((buf array-or-pointer-char)
     (num :int)
     (offset :int))
  :returning :void)

#-(or cmu sbcl)
(def-function ("write_uint" write-uint)
    ((buf array-or-pointer-char)
     (num :unsigned-int)
     (offset :int))
  :returning :void)

#-(or cmu sbcl)
(def-function ("write_float" write-float)
    ((buf array-or-pointer-char)
     (num :float)
     (offset :int))
  :returning :void)

#-(or cmu sbcl)
(def-function ("write_double" write-double)
    ((buf array-or-pointer-char)
     (num :double)
     (offset :int))
  :returning :void)

#-(or cmu sbcl)
(def-function ("offset_charp" offset-char-pointer)
    ((p array-or-pointer-char)
     (offset :int))
  :returning array-or-pointer-char)

;; Allegro and Lispworks use 16-bit unicode characters
#+(or cmu sbcl allegro lispworks)
(defmacro byte-length (s)
  "Return the number of bytes of the internal representation
of a string."
  #+(or lispworks (and allegro ics))
  `(let ((l (length ,s))) (+ l l))
  #-(or lispworks (and allegro ics))
  `(length ,s))

;; for copying the bytes of a string to a foreign buffer
;; memcpy is faster than looping!  For Lispworks this causes
;; a string to array conversion, but I don't know how to do
;; any better (fli:replace-foreign-array is promising?)
#-(or cmu sbcl scl openmcl)
(def-function ("copy_buf" copy-str-to-buf)
    ((dest array-or-pointer-char)
     (dest-offset :int)
     (src array-or-pointer-char)
     (src-offset :int)
     (length :int))
  :returning :void)

#+(or cmu sbcl scl)
(def-function ("copy_buf" copy-str-to-buf)
    ((dest array-or-pointer-char)
     (dest-offset :int)
     (src :cstring)
     (src-offset :int)
     (length :int))
  :returning :void)

;; but OpenMCL can't directly pass string bytes.
#+openmcl
(defun copy-str-to-buf (dest dest-offset src src-offset length)
  "Copy a string to a foreign buffer.  From Gary Byers."
  (declare (optimize (speed 3) (safety 0))
	   (type string src)
	   (type array-or-pointer-char dest)
	   (type fixnum length src-offset dest-offset)
	   (dynamic-extent src dest length))
  (multiple-value-bind (ivector disp)
      (ccl::array-data-and-offset src)
    (ccl::%copy-ivector-to-ptr ivector (+ disp src-offset)
			       dest dest-offset length)))

;; Lisp version, for kicks.  this assumes 8-bit chars!
#+(not (or cmu sbcl scl allegro openmcl lispworks))
(defun copy-str-to-buf (dest dest-offset src src-offset length)
  "Copy a string to a foreign buffer."
  (declare (optimize (speed 3) (safety 0))
	   (type string src)
	   (type array-or-pointer-char dest)
	   (type fixnum length src-offset dest-offset)
	   (dynamic-extent src dest length))
  (typecase src
    (simple-string
     (loop for i fixnum from 0 below length
	   do
	   (setf (deref-array dest 'array-or-pointer-char (+ i dest-offset)) 
		 (char-code (schar src (+ i src-offset))))))
    (string
     (loop for i fixnum from 0 below length
	   do
	   (setf (deref-array dest 'array-or-pointer-char (+ i dest-offset)) 
		 (char-code (char src (+ i src-offset))))))))

;; For copying two foreign buffers
(def-function ("copy_buf" copy-bufs)
    ((dest array-or-pointer-char)
     (dest-offset :int)
     (src array-or-pointer-char)
     (src-offset :int)
     (length :int))
  :returning :void)    


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; buffer-stream functions

(eval-when (:compile-toplevel :load-toplevel)
  (defun process-struct-slot-defs (slot-defs struct)
    (loop for def in slot-defs
	  collect (list (first def) (list (second def) struct)))))

(defmacro with-struct-slots (slot-defs struct &body body)
  `(symbol-macrolet ,(process-struct-slot-defs slot-defs struct)
    ,@body))

(defun resize-buffer-stream (bs length)
  "Resize the underlying buffer of a buffer-stream, copying the old data."
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs)
	   (type fixnum length))
  (with-struct-slots ((buf buffer-stream-buffer)
		      (size buffer-stream-size)
		      (len buffer-stream-length))
    bs		      
    (when (> length len)
      (let ((newlen (max length (* len 2))))
	(declare (type fixnum newlen))
	(let ((newbuf (allocate-foreign-object :char newlen)))
	  ;; technically we just need to copy from position to size.....
	  (copy-bufs newbuf 0 buf 0 size)
	  (free-foreign-object buf)
	  (setf buf newbuf)
	  (setf len newlen)
	  nil)))))

(defun resize-buffer-stream-no-copy (bs length)
  "Resize the underlying buffer of a buffer-stream."
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs)
	   (type fixnum length))
  (with-struct-slots ((buf buffer-stream-buffer)
		      (size buffer-stream-size)
		      (len buffer-stream-length))
    bs		      
    (when (> length len)
      (let ((newlen (max length (* len 2))))
	(declare (type fixnum newlen))
	(let ((newbuf (allocate-foreign-object :char newlen)))
	  (free-foreign-object buf)
	  (setf buf newbuf)
	  (setf len newlen)
	  nil)))))

(defun reset-buffer-stream (bs)
  "'Empty' the buffer-stream."
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs))
  (setf (buffer-stream-size bs) 0)
  (setf (buffer-stream-position bs) 0))

(defun buffer-write-byte (b bs)
  "Write a byte."
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs)
	   (type (unsigned-byte 8) b))
  (with-struct-slots ((buf buffer-stream-buffer)
		      (size buffer-stream-size)
		      (len buffer-stream-length))
    bs		      
    (let ((needed (+ size 1)))
      (when (> needed len)
	(resize-buffer-stream bs needed))
      (setf (deref-array buf '(:array :char) size) b)
      (setf size needed))))

(defun buffer-write-int (i bs)
  "Write a 32-bit signed integer."
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs)
	   (type (signed-byte 32) i))
  (with-struct-slots ((buf buffer-stream-buffer)
		      (size buffer-stream-size)
		      (len buffer-stream-length))
    bs		      
    (let ((needed (+ size 4)))
      (when (> needed len)
	(resize-buffer-stream bs needed))
      (write-int buf i size)
      (setf size needed)
      nil)))

(defun buffer-write-uint (u bs)
  "Write a 32-bit unsigned integer."
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs)
	   (type (unsigned-byte 32) u))
  (with-struct-slots ((buf buffer-stream-buffer)
		      (size buffer-stream-size)
		      (len buffer-stream-length))
    bs		      
    (let ((needed (+ size 4)))
      (when (> needed len)
	(resize-buffer-stream bs needed))
      (write-uint buf u size)
      (setf size needed)
      nil)))

(defun buffer-write-float (d bs)
  "Write a single-float."
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs)
	   (type single-float d))
  (with-struct-slots ((buf buffer-stream-buffer)
		      (size buffer-stream-size)
		      (len buffer-stream-length))
    bs		      
    (let ((needed (+ size 4)))
      (when (> needed len)
	(resize-buffer-stream bs needed))
      (write-float buf d size)
      (setf size needed)
      nil)))

(defun buffer-write-double (d bs)
  "Write a double-float."
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs)
	   (type double-float d))
  (with-struct-slots ((buf buffer-stream-buffer)
		      (size buffer-stream-size)
		      (len buffer-stream-length))
    bs		      
    (let ((needed (+ size 8)))
      (when (> needed len)
	(resize-buffer-stream bs needed))
      (write-double buf d size)
      (setf size needed)
      nil)))

(defun buffer-write-string (s bs)
  "Write the underlying bytes of a string.  On Unicode
Lisps, this is a 16-bit operation."
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs)
	   (type string s))
  (with-struct-slots ((buf buffer-stream-buffer)
		      (size buffer-stream-size)
		      (len buffer-stream-length))
    bs		      
    (let* ((str-bytes (byte-length s))
	   (needed (+ size str-bytes)))
      (declare (type fixnum str-bytes needed)
	       (dynamic-extent str-bytes needed))
      (when (> needed len)
	(resize-buffer-stream bs needed))
      (copy-str-to-buf buf size s 0 str-bytes)
      (setf size needed)
      nil)))

(defun buffer-read-byte (bs)
  "Read a byte."
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs))
  (let ((position (buffer-stream-position bs)))
    (incf (buffer-stream-position bs))
    (deref-array (buffer-stream-buffer bs) '(:array :char) position)))

(defun buffer-read-fixnum (bs)
  "Read a 32-bit signed integer, which is assumed to be a fixnum."
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs))
  (let ((position (buffer-stream-position bs)))
    (setf (buffer-stream-position bs) (+ position 4))
    (the fixnum (read-int (buffer-stream-buffer bs) position))))

(defun buffer-read-int (bs)
  "Read a 32-bit signed integer."
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs))
  (let ((position (buffer-stream-position bs)))
    (setf (buffer-stream-position bs) (+ position 4))
    (the (signed-byte 32) (read-int (buffer-stream-buffer bs) position))))

(defun buffer-read-uint (bs)
  "Read a 32-bit unsigned integer."
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs))
  (let ((position (buffer-stream-position bs)))
    (setf (buffer-stream-position bs) (+ position 4))
    (the (unsigned-byte 32)(read-uint (buffer-stream-buffer bs) position))))

(defun buffer-read-float (bs)
  "Read a single-float."
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs))
  (let ((position (buffer-stream-position bs)))
    (setf (buffer-stream-position bs) (+ position 4))
    (read-float (buffer-stream-buffer bs) position)))

(defun buffer-read-double (bs)
  "Read a double-float."
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs))
  (let ((position (buffer-stream-position bs)))
    (setf (buffer-stream-position bs) (+ position 8))
    (read-double (buffer-stream-buffer bs) position)))

(defun buffer-read-string (bs length)
  "Read a string.  On Unicode Lisps this is a 16-bit operation!"
  (declare (optimize (speed 3) (safety 0))
	   (type buffer-stream bs)
	   (type fixnum length))
  (let ((position (buffer-stream-position bs)))
    (setf (buffer-stream-position bs) (+ position length))
    ;; wide!!!
    #+(and allegro ics)
    (excl:native-to-string 
     (offset-char-pointer (buffer-stream-buffer bs) position) 
     :length length
     :external-format :unicode)
    #+lispworks
    (fli:convert-from-foreign-string 
     (offset-char-pointer (buffer-stream-buffer bs) position)
     :length length :external-format :unicode :null-terminated-p nil)
    #-(or lispworks (and allegro ics))
    (convert-from-foreign-string 
     (offset-char-pointer (buffer-stream-buffer bs) position) 
     :length length :null-terminated-p nil)))

;; Wrapper macro -- handles errno return values
;; makes flags into keywords
;; makes keyword args, cstring wrappers


(eval-when (:compile-toplevel)
  (defun make-wrapper-args (args flags keys)
    (if (or flags keys)
	(append (remove-keys (remove 'flags args) keys)
		`(&key ,@flags ,@keys))
	(remove 'flags args)))
  
  (defun remove-keys (args keys)
    (if keys
	(loop for key in keys
	      for kw = (if (atom key) key (first key))
	      for wrapper-args = (remove kw args) then (remove kw wrapper-args)
	      finally (return wrapper-args))
	args))
  
  (defun make-fun-args (args flags)
    (if flags
	(substitute (cons 'flags (symbols-to-kw-pairs flags)) 'flags args)
	(substitute 0 'flags args)))
  
  (defun make-out-args (count)
    (loop for i from 1 to count
	  collect (gensym)))
  
  (defun symbols-to-kw-pairs (symbols)
    (loop for symbol in symbols
	  append (list (intern (symbol-name symbol) "KEYWORD")
		       symbol)))
  
  (defun symbols-to-pairs (symbols)
    (loop for symbol in symbols
	  collect (list symbol symbol)))
  )

(defmacro wrap-errno (names args &key (keys nil) (flags nil)
		      (cstrings nil) (outs 1) (declarations nil)
		      (documentation nil)
		      (transaction nil))
  (let ((wname (if (listp names) (first names) names))
	(fname (if (listp names) (second names) 
		   (intern (concatenate 'string "%" (symbol-name names)))))
	(wrapper-args (make-wrapper-args args flags keys))
	(fun-args (make-fun-args args flags))
	(errno (gensym)))
    (if (> outs 1)
	(let ((out-args (make-out-args outs)))
	  `(defun ,wname ,wrapper-args
	    ,@(if documentation (list documentation) (values))
	    ,@(if declarations (list declarations) (values))	    
	    (with-cstrings ,(symbols-to-pairs cstrings)
	      (multiple-value-bind ,out-args
		  (,fname ,@fun-args)
		(let ((,errno ,(first out-args)))
		  (declare (type fixnum ,errno))
		  (cond
		    ((= ,errno 0) (values ,@(rest out-args)))
		    ,@(if transaction
			  (list `((or (= ,errno DB_LOCK_DEADLOCK)
				   (= ,errno DB_LOCK_NOTGRANTED))
				  (throw 'transaction ,transaction)))
			  (values))
		    (t (error 'db-error :errno ,errno))))))))
	`(defun ,wname ,wrapper-args
	  ,@(if documentation (list documentation) (values))
	  ,@(if declarations (list declarations) (values))
	  (with-cstrings ,(symbols-to-pairs cstrings)
	    (let ((,errno (,fname ,@fun-args)))
	      (declare (type fixnum ,errno))
	      (cond 
		((= ,errno 0) nil)
		,@(if transaction
		      (list `((or (= ,errno DB_LOCK_DEADLOCK)
			       (= ,errno DB_LOCK_NOTGRANTED))
			      (throw 'transaction ,transaction)))
		      (values))
		(t (error 'db-error :errno ,errno)))))))))

(defmacro flags (&key auto-commit joinenv init-cdb init-lock init-log
		 init-mpool init-rep init-txn recover recover-fatal lockdown
		 private system-mem thread force dirty-read create excl nommap
		 rdonly truncate txn-nosync txn-nowait txn-sync lock-nowait
		 dup dup-sort current first get-both get-both-range last next
		 next-dup next-nodup prev prev-nodup set set-range
		 after before keyfirst keylast
		 no-dup-data no-overwrite nosync position set-lock-timeout
		 set-transaction-timeout)
  (let ((flags (gensym)))
    `(let ((,flags 0))
      (declare (type fixnum ,flags))
      ,@(when auto-commit `((when ,auto-commit (setq ,flags (logior ,flags DB_AUTO_COMMIT)))))
      ,@(when joinenv `((when ,joinenv (setq ,flags (logior ,flags DB_JOINENV)))))
      ,@(when init-cdb `((when ,init-cdb (setq ,flags (logior ,flags DB_INIT_CDB)))))
      ,@(when init-lock `((when ,init-lock (setq ,flags (logior ,flags DB_INIT_LOCK)))))
      ,@(when init-log `((when ,init-log (setq ,flags (logior ,flags DB_INIT_LOG)))))
      ,@(when init-mpool `((when ,init-mpool (setq ,flags (logior ,flags DB_INIT_MPOOL)))))
      ,@(when init-rep `((when ,init-rep (setq ,flags (logior ,flags DB_INIT_REP)))))
      ,@(when init-txn `((when ,init-txn (setq ,flags (logior ,flags DB_INIT_TXN)))))
      ,@(when recover `((when ,recover (setq ,flags (logior ,flags DB_RECOVER)))))
      ,@(when recover-fatal `((when ,recover-fatal (setq ,flags (logior ,flags DB_RECOVER_FATAL)))))
      ,@(when lockdown `((when ,lockdown (setq ,flags (logior ,flags DB_LOCKDOWN)))))
      ,@(when private `((when ,private (setq ,flags (logior ,flags DB_PRIVATE)))))
      ,@(when system-mem `((when ,system-mem (setq ,flags (logior ,flags DB_SYSTEM_MEM)))))
      ,@(when thread `((when ,thread (setq ,flags (logior ,flags DB_THREAD)))))
      ,@(when force `((when ,force (setq ,flags (logior ,flags DB_FORCE)))))
      ,@(when dirty-read `((when ,dirty-read (setq ,flags (logior ,flags DB_DIRTY_READ)))))
      ,@(when create `((when ,create (setq ,flags (logior ,flags DB_CREATE)))))
      ,@(when excl `((when ,excl (setq ,flags (logior ,flags DB_EXCL)))))
      ,@(when nommap `((when ,nommap (setq ,flags (logior ,flags DB_NOMMAP)))))
      ,@(when rdonly `((when ,rdonly (setq ,flags (logior ,flags DB_RDONLY)))))
      ,@(when truncate `((when ,truncate (setq ,flags (logior ,flags DB_TRUNCATE)))))
      ,@(when txn-nosync `((when ,txn-nosync (setq ,flags (logior ,flags DB_TXN_NOSYNC)))))
      ,@(when txn-nowait `((when ,txn-nowait (setq ,flags (logior ,flags DB_TXN_NOWAIT)))))
      ,@(when txn-sync `((when ,txn-sync (setq ,flags (logior ,flags DB_TXN_SYNC)))))
      ,@(when lock-nowait `((when ,lock-nowait (setq ,flags (logior ,flags DB_LOCK_NOWAIT)))))
      ,@(when dup `((when ,dup (setq ,flags (logior ,flags DB_DUP)))))
      ,@(when dup-sort `((when ,dup-sort (setq ,flags (logior ,flags DB_DUPSORT)))))
      ,@(when current `((when ,current (setq ,flags (logior ,flags DB_CURRENT)))))
      ,@(when first `((when ,first (setq ,flags (logior ,flags DB_FIRST)))))
      ,@(when get-both `((when ,get-both (setq ,flags (logior ,flags DB_GET_BOTH)))))
      ,@(when get-both-range `((when ,get-both-range (setq ,flags (logior ,flags DB_GET_BOTH_RANGE)))))
      ,@(when last `((when ,last (setq ,flags (logior ,flags DB_LAST)))))
      ,@(when next `((when ,next (setq ,flags (logior ,flags DB_NEXT)))))
      ,@(when next-dup `((when ,next-dup (setq ,flags (logior ,flags DB_NEXT_DUP)))))
      ,@(when next-nodup `((when ,next-nodup (setq ,flags (logior ,flags DB_NEXT_NODUP)))))
      ,@(when prev `((when ,prev (setq ,flags (logior ,flags DB_PREV)))))
      ,@(when prev-nodup `((when ,prev-nodup (setq ,flags (logior ,flags DB_PREV_NODUP)))))
      ,@(when set `((when ,set (setq ,flags (logior ,flags DB_SET)))))
      ,@(when set-range `((when ,set-range (setq ,flags (logior ,flags DB_SET_RANGE)))))
      ,@(when after `((when ,after (setq ,flags (logior ,flags DB_AFTER)))))
      ,@(when before `((when ,before (setq ,flags (logior ,flags DB_BEFORE)))))
      ,@(when keyfirst `((when ,keyfirst (setq ,flags (logior ,flags DB_KEYFIRST)))))
      ,@(when keylast `((when ,keylast (setq ,flags (logior ,flags DB_KEYLAST)))))
      ,@(when no-dup-data `((when ,no-dup-data (setq ,flags (logior ,flags DB_NODUPDATA)))))
      ,@(when no-overwrite `((when ,no-overwrite (setq ,flags (logior ,flags DB_NOOVERWRITE)))))
      ,@(when nosync `((when ,nosync (setq ,flags (logior ,flags DB_NOSYNC)))))
      ,@(when position `((when ,position (setq ,flags (logior ,flags DB_POSITION)))))    
      ,@(when set-lock-timeout `((when ,set-lock-timeout (setq ,flags (logior ,flags DB_SET_LOCK_TIMEOUT)))))
      ,@(when set-transaction-timeout `((when ,set-transaction-timeout (setq ,flags (logior ,flags DB_SET_TXN_TIMEOUT)))))
      ,flags)))


