
(in-package :ele-tests)

(deftest testoid
    (progn
      (ele::next-oid *store-controller*)
      (let ((oid (ele::next-oid *store-controller*)))
	(with-open-store (*testdb-path*)
	  (< oid (ele::next-oid *store-controller*)))))
  t)

(defclass blob ()
  ((slot1 :accessor slot1 :initarg :slot1)
   (slot2 :accessor slot2 :initarg :slot2)))

(defvar keys (loop for i from 1 to 1000 
		   collect (concatenate 'string "key-" (prin1-to-string i))))

(defvar objs (loop for i from 1 to 1000
		   collect (make-instance 'blob
					  :slot1 i
					  :slot2 (* i 100))))

(defvar bt)

(deftest btree-make
    (finishes (setq bt (make-instance 'btree)))
  t)

(setq *auto-commit* nil)

(deftest btree-put
    (finishes
      (with-transaction ()
	(loop for obj in objs
	      for key in keys
	      do (setf (get-value key bt) obj))))
  t)

(deftest btree-get
    (loop for key in keys
	  for i from 1 to 1000
	  for obj = (get-value key bt)
	  always
	  (and (= (slot1 obj) i)
	       (= (slot2 obj) (* i 100))))
  t)

(deftest remove-kv
    (finishes (with-transaction () (remove-kv "key-1" bt)))
  t)

(deftest removed
    (not (get-value "key-1" bt))
  t)

(deftest map-btree
    (let ((ks nil)
	  (vs nil))
      (flet ((mapper (k v) (push k ks) (push v vs)))
	(map-btree #'mapper bt))
      (values
       (and (subsetp ks (cdr keys) :test #'equalp) 
	    (subsetp (cdr keys) ks :test #'equalp))))
  t)

(defvar indexed)
(defvar index1)
(defvar index2)

(deftest indexed-btree-make
    (finishes (with-transaction ()
		(setq indexed (make-instance 'indexed-btree))))
  t)

(defun key-maker (s key value)
  (declare (ignore s key))
  (values t (slot1 value)))

(deftest add-indices
    (finishes
     (with-transaction ()
       (setf index1
	     (add-index indexed :index-name 'slot1 :key-form 'key-maker))
       (setf index2
	     (add-index indexed :index-name 'slot2
			:key-form '(lambda (s key value) 
				    (declare (ignore s key))
				    (values t (slot2 value)))))))
  t)

(deftest test-indices
    (values
     (= (hash-table-count (indices indexed)) 2)
     (eq index1 (gethash 'slot1 (indices indexed)))
     (eq index2 (gethash 'slot2 (indices indexed))))
  t t t)

#|
(deftest safe-indexed-put
    (finishes
     (loop for i from 1 to 1000
	   for obj in objs
	   for key in keys
	   do
	   (setf (get-value key indexed) obj)
	   (loop for j from 1 to i
		 for key2 in keys
		 for obj2 = (get-value key2 indexed)
		 always
		 (and (= (slot1 obj2) j)
		      (= (slot2 obj2) (* j 100))))))
  t) |#
   
(deftest indexed-put
    (finishes
      (with-transaction ()
	(loop for obj in objs
	      for key in keys
	      do (setf (get-value key indexed) obj))))
  t)

(deftest indexed-get
    (loop for key in keys
	  for i from 1 to 1000
	  for obj = (get-value key indexed)
	  always 
	  (and (= (slot1 obj) i)
	       (= (slot2 obj) (* i 100))))
  t)

(deftest indexed-get-from-slot1
    (loop with index = (get-index indexed 'slot1)
	  for i from 1 to 1000
	  for obj = (get-value i index)
	  always
	  (= (slot1 obj) i))
  t)
	  
(deftest indexed-get-from-slot2
    (loop with index = (get-index indexed 'slot2)
	  for i from 1 to 1000
	  for obj = (get-value (* i 100) index)
	  always
	  (= (slot2 obj) (* i 100)))
  t)
	  
(deftest remove-kv-indexed
    (finishes (remove-kv "key-1" indexed))
  t)

(deftest no-key-nor-indices
    (values
     (get-value "key-1" indexed)
     (get-primary-key 1 index1)
     (get-primary-key 100 index2))
  nil nil nil)

(deftest remove-kv-from-slot1
    (finishes (remove-kv 2 index1))
  t)

(deftest no-key-nor-indices-slot1
    (values
     (get-value "key-2" indexed)
     (get-primary-key 2 index1)
     (get-primary-key 200 index2))
  nil nil nil)

(deftest remove-kv-from-slot2
    (finishes (remove-kv 300 index2))
  t)

(deftest no-key-nor-indices-slot2
    (values
     (get-value "key-3" indexed)
     (get-primary-key 3 index1)
     (get-primary-key 300 index2))
  nil nil nil)

(deftest map-indexed
    (let ((ks nil)
	  (vs nil))
      (flet ((mapper (k v) (push k ks) (push v vs)))
	(map-btree #'mapper indexed))
      (values
       (and (subsetp ks (cdddr keys) :test #'equalp) 
	    (subsetp (cdddr keys) ks :test #'equalp))))
  t)

(deftest get-first
    (with-transaction ()
      (with-btree-cursor (c index1)
	(multiple-value-bind (has k v)
	    (cursor-first c)
	  (declare (ignore has v))
	  (= k 4))))
  t)

(deftest get-first2
    (with-transaction ()
      (with-btree-cursor (c index2)
	(multiple-value-bind (has k v)
	    (cursor-first c)
	  (declare (ignore has v))
	  (= k 400))))
  t)

(deftest get-last
    (with-transaction ()
      (with-btree-cursor (c index1)
	(multiple-value-bind (has k v)
	    (cursor-last c)
	  (declare (ignore has v))
	  (= k 1000))))
  t)

(deftest get-last2
    (with-transaction ()
      (with-btree-cursor (c index2)
	(multiple-value-bind (has k v)
	    (cursor-last c)
	  (declare (ignore has v))
	  (= k 100000))))
  t)

(deftest set
    (with-transaction ()
      (with-btree-cursor (c index1)
	(multiple-value-bind (has k v)
	    (cursor-set c 200)
	  (declare (ignore has k))
	  (= (slot1 v) 200))))
  t)

(deftest set2
    (with-transaction ()
      (with-btree-cursor (c index2)
	(multiple-value-bind (has k v)
	    (cursor-set c 500)
	  (declare (ignore has k))
	  (= (slot2 v) 500))))
  t)

(deftest set-range
    (with-transaction ()
      (with-btree-cursor (c index1)
	(multiple-value-bind (has k v)
	    (cursor-set-range c 199.5)
	  (declare (ignore has k))
	  (= (slot1 v) 200))))
  t)

(deftest set-range2
    (with-transaction ()
      (with-btree-cursor (c index2)
	(multiple-value-bind (has k v)
	    (cursor-set-range c 501)
	  (declare (ignore has k))
	  (= (slot2 v) 600))))
  t)

(defvar indexed2)
(defvar index3)

(deftest make-indexed2
    (finishes (with-transaction ()
		(setq indexed2 (make-instance 'indexed-btree))))
  t)

(defun crunch (s k v)
  (declare (ignore s v))
  (values t (floor (/ k 10))))

(deftest add-indices2
    (finishes
      (with-transaction () 
	(setq index3
	      (add-index indexed2 :index-name 'crunch :key-form 'crunch))))
  t)

(deftest put-indexed2
    (finishes
      (with-transaction () 
	(loop for i from 0 to 10000
	      do
	      (setf (get-value i indexed2) (- i)))))
  t)

(deftest get-indexed2
    (loop for i from 0 to 10000
	  always (= (- i) (get-value i indexed2)))
  t)

(deftest get-from-index3
    (loop for i from 0 to 1000
	  always (= (* i -10) (get-value i index3)))
  t)
	  

(deftest dup-test
    (with-transaction ()
      (with-btree-cursor (curs index3)
	(loop for (more k v) = (multiple-value-list
				(cursor-first curs))
	      then (multiple-value-list (cursor-next-dup curs))
	      while more
	      collect v)))
  (0 -1 -2 -3 -4 -5 -6 -7 -8 -9))
	      

(deftest nodup-test
    (with-transaction ()
      (with-btree-cursor (curs index3)
	(loop for (m k v) = (multiple-value-list (cursor-next-nodup curs))
	      for i from 0 downto -9990 by 10
	      while m
	      always (= v i))))
  t)

(deftest prev-nodup-test
    (with-transaction ()
      (with-btree-cursor (curs index3)
	(cursor-last curs)
	(loop for (m k v) = (multiple-value-list (cursor-prev-nodup curs))
	      for i from -9999 to -9 by 10
	      while m
	      always (= v i))))
  t)

(deftest pnodup-test
    (with-transaction ()
      (with-btree-cursor (curs index3)
	(loop for (m k v p) = (multiple-value-list (cursor-pnext-nodup curs))
	      for i from 0 to 9990 by 10
	      while m
	      always (= p i))))
  t)

(deftest pprev-nodup-test
    (with-transaction ()
      (with-btree-cursor (curs index3)
	(cursor-last curs)
	(loop for (m k v p) = (multiple-value-list (cursor-pprev-nodup curs))
	      for i from 9999 downto 9 by 10
	      while m
	      always (= p i))))
  t)

(deftest indexed-delete 
    (finishes
      (with-transaction ()
	(with-btree-cursor (curs index3)
	  (cursor-last curs)
	  (cursor-delete curs))))
  t)

(deftest test-deleted
    (values
     (get-value 10000 indexed2)
     (get-value 1000 index3))
  nil nil)
		      
(deftest indexed-delete2
    (finishes
      (with-transaction ()
	(with-btree-cursor (curs index3)
	  (cursor-first curs)
	  (cursor-next-dup curs)
	  (cursor-delete curs))))
  t)

(deftest test-deleted2
    (values
     (get-value 0 indexed2)
     (get-value 0 index3)
     (get-value 1 indexed2)
     (with-btree-cursor (c index3)
       (cursor-first c)
       (multiple-value-bind (m k v) (cursor-next c)
	 v)))
  0 0 nil -2)

(deftest get-both 
    (with-btree-cursor (c indexed2)
      (cursor-get-both c 200 -200))
  t 200 -200)

(deftest pget-both 
    (with-btree-cursor (c index3)
      (multiple-value-bind (m k v p)
	  (cursor-pget-both c 10 107)
	(values k v p)))
  10 -107 107)

(deftest pget-both-range
    (with-btree-cursor (c index3)
      (multiple-value-bind (m k v p)
	  (cursor-pget-both-range c 10 106.5)
	(values k v p)))
  10 -107 107)

(defmacro pcursor-pkey (form)
  `(multiple-value-bind (m k v p)
    ,form
    (declare (ignore m k v))
    p))

(deftest pcursor
    (with-btree-cursor (c index3)
      (values
       (pcursor-pkey (cursor-pfirst c))
       (pcursor-pkey (cursor-pnext c))
       (pcursor-pkey (cursor-pnext-nodup c))
       (pcursor-pkey (cursor-pnext-dup c))
       (pcursor-pkey (cursor-pprev c))
       (pcursor-pkey (cursor-pprev-nodup c))
       (pcursor-pkey (cursor-plast c))
       (pcursor-pkey (cursor-pset c 300))
       (pcursor-pkey (cursor-pset-range c 199.5))
       (pcursor-pkey (cursor-pget-both c 10 101))
       (pcursor-pkey (cursor-pget-both-range c 11 111.4))))
      
  0 2 10 11 10 9 9999 3000 2000 101 112)

(defvar index4)

(deftest newindex
    (finishes
     (with-transaction () 
       (setq index4
	     (add-index indexed2 :index-name 'crunch :key-form 'crunch
			:populate t))))
  t)

(deftest pcursor2
    (with-btree-cursor (c index4)
      (values
       (pcursor-pkey (cursor-pfirst c))
       (pcursor-pkey (cursor-pnext c))
       (pcursor-pkey (cursor-pnext-nodup c))
       (pcursor-pkey (cursor-pnext-dup c))
       (pcursor-pkey (cursor-pprev c))
       (pcursor-pkey (cursor-pprev-nodup c))
       (pcursor-pkey (cursor-plast c))
       (pcursor-pkey (cursor-pset c 300))
       (pcursor-pkey (cursor-pset-range c 199.5))
       (pcursor-pkey (cursor-pget-both c 10 101))
       (pcursor-pkey (cursor-pget-both-range c 11 111.4))))
      
  0 2 10 11 10 9 9999 3000 2000 101 112)
