
(in-package "ELE-TESTS")
(use-package "SLEEPYCAT")

(defvar env)
(defvar db)
(defvar keys)

(defun make-keys (iters)
  (loop for i from 1 to iters
	collect (concatenate 'string "key-" (prin1-to-string i))))

(setq keys (make-keys 1000))

(defun prepare()
  (setq env (db-env-create))
  (db-env-open env "test" :create t :init-txn t :init-lock t 
	       :init-mpool t :init-log t :thread t
	       :recover-fatal t)
  
  (setq db (db-create env))
  (db-open db :file "foo" :database "bar" :type DB-BTREE
	   :auto-commit t :create t :thread t))

(deftest prepares
    (finishes (prepare)) t)

(deftest put-alot
    (finishes
     (loop for key in keys
	   do
	   (db-put db key key :auto-commit t)))
  t)
  
(defun get-alot ()
  (loop for key in keys
	always (string= key (db-get db key))))

(deftest put-right (get-alot) t)

(deftest put-alot-b 
    (finishes
     (with-transaction (:environment env)
       (loop for key in keys
	     do
	     (db-put db key key))))
  t)

(deftest put-right-b (get-alot) t)

(defun txn-alot (iters)
  (loop for i from 1 to iters
	do
	(with-transaction (:environment env)
	  (db-put db "mykey" "mydatum"))))	     

(defun get-alot-b (keys)
  (loop for key in keys
	do
	(db-get-buffered db key)))

(defun foreign-test (ln iters)
  (with-transaction (:environment env)
    (loop for i fixnum from 1 to iters
	  with write-buf of-type array-or-pointer-char = (uffi:allocate-foreign-object :char ln)
	  with str string = (make-string ln :initial-element #\c)
	  with key-buf of-type array-or-pointer-char = (uffi:allocate-foreign-object :char 2)
	  do
	  (copy-str-to-buf "fs" key-buf)
	  (copy-str-to-buf str write-buf)
	  (db-put-buffered db key-buf 2 write-buf ln)
	  finally 
	  (progn 
	    (uffi:free-foreign-object write-buf)
	    (uffi:free-foreign-object key-buf)))))

(defun cstring-test (ln iters)
  (with-transaction (:environment env)
    (loop for i fixnum from 1 to iters
	  with str string = (make-string ln :initial-element #\c)
	  do
	  (db-put db "fs" str))))

(defun cleanup ()
  (db-close db)
  
  (db-env-dbremove env "foo" :database "bar")
  (db-env-close env)
  (setq env (db-env-create))
  (db-env-remove env "test"))

